#!/bin/bash
# Last update: 14/08/2020
# by Emanuele Tomasi aka targzeta <targzeta@gmail.com>

# Copyright (C) 2011-2012 Emanuele Tomasi aka targzeta <targzeta@gmail.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# Scova i file presenti nella root directory ma che però non sono stati
# installati da nessuno dei pacchetti Slackware.

# -h for help

SCRIPT_NAME=${0##*/} # Nome dello script
SCRIPT_AUTHOR='Emanuele Tomasi aka targzeta <targzeta@gmail.com>'
SCRIPT_HOMEPAGE='http://www.slacky.eu/wikislack/index.php?title=Find_external_SlackPkg_files'
SCRIPT_VERSION=1.2

## PERSONALIZZAZIONE ##

# Il file con l'elenco dei path da ignorare. Se il path si riferisce
# ad un file, allora il file sarà ignorato, se invece si tratta di una
# directory, questa non sarà scandita, con la conseguenza che tutti i
# file ivi presenti saranno ignorati.
#
# Sono ignorate le righe vuote e le righe che iniziano con il '#'. Il
# file deve contenere un path per ogni riga, eventualmente si può
# specificare un comando da controllare, in questo caso il path sarà
# ignorato solo se esiste tale comando. Il comando deve essere
# separato dal path da almeno un carattere di spaziatura. Ad esempio,
# se il file di filtro contiene le seguenti righe:
# ---
# # My comment
#
# /home
# /usr/share/fonts/*.dir
# /usr/share/mime                 update-mime-database
# ---
# Allora lo script ignorerà:
# * la directory /home
# * tutti i file generati dal comando 'mkfontdir'
# * la directory /usr/share/mime se, e solo se, esiste il comando
#   'update-mime-database'
#
# Nota: i path sono passati al comando 'find' tramite l'uso della sua
# opzione '-path'. Si veda il manuale di 'find' per conoscere il
# formato accettato.
#
# Nota: se il path da filtrare contiene uno spazio, allora lo script
# penserà che ciò che si trova dopo lo spazia sia un comando da
# controllare. In questo caso, affinchè tutto funzioni, bisogna
# aggiungere un altro spazio e poi un comando sicuramente esistente
# (ad esempio 'bash', visto che questo script viene eseguito proprio
# dalla bash).
#
# Nota: altri path saranno comunque ignorati, si veda la funzione
# filter_system_path().
FILTER_FILE=~/.find_external_SlackPkg_files

## FINE PERSONALIZZAZIONE ##

# Ritorna l'elenco dei path di sistema da ignorare. Queste directory
# non vanno scandite.
function _filter_system_path()
{
  cat <<EOF
/dev
/proc
/root
/sys
/var/lib/pkgtools
/var/log/packages
/var/log/scripts
/var/log/removed_packages
/var/log/removed_scripts
EOF
}

# Ritorna l'elenco dei path utente da ignorare controllando che
# l'eventuale programma settato come condizione esiste.
function _user_path_filtered
{
  if [ -f $FILTER_FILE ]
  then
    (IFS=$'\n'
     for line in \
       $(sed '/^[[:space:]]*#/d;/^[[:space:]]$/d' ${FILTER_FILE})
     do
       path=$(echo $line | \
                sed 's/\(.*[^[:space:]]\)[[:space:]]\+[^[:space:]]\+/\1/')
       program=$(echo $line | \
                   sed 's/.*[[:space:]]\([^[:space:]]\+\)/\1/')

       if [[ $path == $program ]] || /bin/which $program >& /dev/null
       then
         echo $path
       fi
     done
    )
  fi
}

# Stampa l'argomento $1 seguito dalle informazioni per eseguire l'help.
# Poi ritorna uno stato di uscita pari a $2
function _exit
{
  echo -e "${1}\n\n${SCRIPT_NAME} -h for the help"
  exit $2
}

# Stampa l'help dello script
function _help
{
  cat <<EOS
Usage: ${SCRIPT_NAME} [option]
Finds those files that aren't installed by any Slackware package.

Options:
  -f    Shows the filtered paths. Paths that will be ignored.
  -h    Display this help and exit
  -v    Display version and exit
EOS
}

# Stampa la versione
function _version
{
  cat <<EOS
$SCRIPT_NAME $SCRIPT_VERSION
Copyright © 2020 ${SCRIPT_AUTHOR}
License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.

Homepage: ${SCRIPT_HOMEPAGE}
Written by ${SCRIPT_AUTHOR%<*}
EOS
}

# Controlla l'esistenza di tutti i programmi dati i loro nomi come argomento.
# Se un programma non viene trovato, allora, stampa il nome del o dei programmi
# mancanti ed esce. Si usa il programma which per controllare che esistono i
# programmi passati. Si assume che which sia in /bin/.
function _command_exists
{
  local error_string=''

  [ ! -x /bin/which ] && _exit "which: no such command" 1

  for progr in $@
  do
    if ! /bin/which $progr >& /dev/null
    then
      [ -n "$error_string" ] && error_string+="\n"
      error_string+="${progr}: no such command"
    fi
  done

  [ -n "$error_string" ] && _exit "$error_string" 1
}

# Controllo degli argomenti dello script in cerca di opzioni
enable getopts echo exit let local shift

# Impostazione del separatore di campi al valore di default
IFS=$'\n\t '

# Controllo dell'esistenza di tutti i comandi necessari
_command_exists awk comm find mktemp realpath rm sed sort uniq

FLAG_F=0
while getopts :fhv opzione
do
  case $opzione in
    f)
      FLAG_F=1
      ;;

    h)
      _help
      exit 0
      ;;
    v)
      _version
      exit 0
      ;;
    ?)
    _exit "-${OPTARG} : invalid argument." 1
    ;;
  esac
done

# Shift degli argomenti in modo da avere $1 che punta al primo argomento
# che non è una opzione
while (( $OPTIND != 1 ))
do
  shift
  let --OPTIND
done

### MAIN ###
first=0
for path in $(_filter_system_path) $(_user_path_filtered)
do
  if (( $FLAG_F == 1 ))
  then
    echo $path
  elif (( $first == 0 ))
  then
    path_to_prune="-path $path"
    first=1
  else
    path_to_prune="$path_to_prune -o -path $path"
  fi
done
(( $FLAG_F == 1 )) && exit 0

# Creo la directory temporanea di lavoro
! WORK_DIR=$(mktemp -d -p /tmp) && exit 1

# Mi assicuro di eliminare la directory temporanea all'uscita
trap "rm -r $WORK_DIR" EXIT

cd $WORK_DIR

# Scandiamo tutta la root directory saltando i path filtrati
find / -mindepth 1 \( $path_to_prune \) -prune \
     -o \( -type d -printf "%p/\n" \) -o -print | sort >  found_files.s

# Iniziamo a creare la lista dei file installati dai pacchetti...
# Per ogni file di info trovato sotto /var/log/packages/ bisogna:
# - saltare la parte introduttiva di descrizione del pacchetto;
# - saltare la directory install/;
# - inserire uno slash iniziale ai file;
# - se il file finisce con '.new', aggiungerci anche il relativo file
#   senza '.new'.
sed -s -n '\|^\./|b next;d;:next n;\|^install/|b next;s|^|/|;p;
           /\.new$/{s/\.new$//;p};b next' /var/log/packages/* > installed_files

# ...aggiungiamo i link simbolici creati dai doinst.sh dei pacchetti...
# Nota: aggiungiamo anche i path assoluti linkati, per sistemare problemi come
# '/bin/bash' e '/etc/localtime', creati da doinst.sh complessi.
grep -h '^[^#]*cd [^ ]* ; *ln ' /var/log/scripts/* |\
  while read _ _ dir _ _ _ target link _
  do
    echo /${dir%/}/$link
    [ ${target:0:1} == '/' ] && echo $target
  done | sort -u > symbolic_links.s
cat symbolic_links.s >> installed_files

# ...aggiungiamo tutte le librerie comprese nelle glibc...
# I pacchetti glibc-<VERSION> o glibc-solib-<VERSION> installano le librerie
# nella directory 'incoming' e poi le spostano con il doinst.sh. Aggiungiamo
# nuovamente questi file senza la componente '/incoming/'.
# Nota: si potrebbe rimuovere la componente '/incoming/' direttamente nel file
# 'installed_files', ma così facendo si rischierebbe di rimuovere anche file
# installati da altri pacchetti.
sed -n '\|/incoming/[^ ]|{s|^|/|;s|/incoming/|/|;p}' \
    /var/log/packages/glibc-* >> installed_files

# Scoviamo i file della root directory che non sono stati installati da nessun
# pacchetto. Ovvero, tutti quelli prensenti in 'foud_files.s' e non in
# 'installed_files.s'.
sort -u installed_files > installed_files.s
comm -23 --nocheck-order found_files.s installed_files.s > temp_output0.s

# Si compattano le directory.
#
# Nota, il file 'temp_output0.s' deve essere ordinato, in modo che le
# directory vengano prima dei file ivi contenuti.
cat temp_output0.s |\
  awk 'BEGIN { row = "" }
      {
        if (row == "" || ! match(row, "/$") || ! match($0, "^" row)) {
          print;
          row = $0
        }
      }' > temp_output1.s

# Si filtrano i link simbolici che puntano a file installati.
#
# Molti link simbolici sono creati a runtime e, anche se non installati da
# nessun pacchetto, vanno comunque considerati come corretti.
cat temp_output1.s |\
  while read file
  do
    [ -h "$file" -a -e "$file" ] && ! grep "^$(realpath -m '$file')$" temp_output0.s && continue
    echo "$file"
  done

exit 0
